/*
 *  Tetris-like Game Example -- WiFiBoy32 Demo Code for Arduino/ESP32
 *  
 *  Oct 26, 2017 Created. (derek@wifiboy.org)
 *
 */
#include <wifiboy32.h>
#include <Preferences.h>

uint8_t blk_t[28][4] = { // seven shapes, four rotation types
  {0,10,20,30},{0,1,2,3},{0,10,20,30},{0,1,2,3},
  {1,10,11,12},{0,10,11,20},{0,1,2,11},{1,10,11,21},
  {0,1,10,11},{0,1,10,11},{0,1,10,11},{0,1,10,11},
  {0,10,20, 21},{0,1,2,10},{0,1,11,21},{2,10,11,12},
  {1,11,20,21},{0,10,11,12},{0,1,10,20},{0,1,2,12},
  {0,1,11,12},{1,10,11,20},{0,1,11,12},{1,10,11,20},
  {1,2,10,11},{0,10,11,21},{1,2,10,11},{0,10,11,21}
};

uint16_t bcolor[8]={
  wbBLACK, wbBLUE, wbRED, wbGREEN, wbCYAN, wbMAGENTA, wbYELLOW, wbWHITE
};

// we use a lot of dirty globals for efficiency
Preferences preferences;
uint8_t board[20][10], offboard[20][10]; // keep an off board for dirty-rect-rendering
uint16_t cx, cy, kc, task_c, sfx_on, sfxn, sfxc, freq, fall_time, fall_limit, stage_limit,highscore; 
uint16_t ctype, nctype, level, pts, pn, i, j, k, pos, px, py, key, last_key, cline;
char tmpbuf[10];
int rot, ret;

void init_keys()
{
  pinMode(17,INPUT); pinMode(32,INPUT); pinMode(33,INPUT); pinMode(27,INPUT);
  pinMode(39,INPUT); pinMode(23,INPUT); pinMode(34,INPUT); pinMode(35,INPUT);
  kc = 0;
}

uint16_t getkey() // global key: 8bit key value: L R U D A B Sel Start 
{
  return(key = digitalRead(17)*128+digitalRead(32)*64+digitalRead(33)*32+digitalRead(27)*16
        +digitalRead(35)*8+digitalRead(34)*4+digitalRead(39)*2+digitalRead(23)*1);
}

void draw_board()
{
  for(i=0; i<20; i++)
    for(j=0; j<10; j++)
      if (offboard[i][j]!=board[i][j]) // dirty rectangle check
        wb32_fillRect(j*16+1, i*16, 15, 15, bcolor[offboard[i][j]=board[i][j]]);
}

void clear_board()
{
  for(i=0; i<20; i++)     // board is 10x20
    for(j=0; j<10; j++) {
      offboard[i][j]=255; // initial state
      board[i][j]=0;      // clear all blocks
    }
  wb32_fillRect(1, 0, 159, 319, wbBLACK);
}

void get_pos(uint16_t pos)
{
  py=pos/10; px = pos%10;
}

void draw_blocks(uint16_t x, uint16_t y, uint16_t v)
{
  for(int i=0; i<4; i++) {
    get_pos(y*10 + x + blk_t[ctype*4+rot][i]); 
    board[py][px]=v; 
  }
}

void draw_preview() // next blocks
{
  wb32_fillRect(175, 30, 56, 56, wbBLACK);
  for(i=0; i<4; i++) {
    get_pos(blk_t[nctype*4][i]);
    wb32_fillRect(175+px*13, 30+py*13, 11, 11, bcolor[nctype+1]);
  }
}

void update_cline_level()
{
  if (cline > 150) level = 10;
  else if (cline > 120) level=9;
  else if (cline > 100) level=8;
  else if (cline > 80) level=7;
  else if (cline > 60) level=6;
  else if (cline > 40) level=5;
  else if (cline > 25) level=4;
  else if (cline > 15) level=3;
  else if (cline > 10) level=2;
  else if (cline > 5) level=1;
}

void update_score()
{
  wb32_setTextColor(0xffff, 0xffff);
  wb32_fillRect(175, 230, 50, 15, wbBLACK);
  wb32_drawString(itoa(cline, tmpbuf, 10), 175, 230, 1, 2);
  update_cline_level();
  fall_limit = stage_limit = (level<=12?13-level:1);
  wb32_fillRect(175, 285, 50, 15, wbBLACK);
  wb32_drawString(itoa(level+1, tmpbuf, 10), 175, 285, 1, 2);
  wb32_fillRect(175, 175, 50, 15, wbBLACK);
  wb32_drawString(itoa(pts, tmpbuf, 10), 175, 175, 1, 2);
  if (pts>highscore) {
    highscore=pts;
    preferences.begin("tetris",false);
    preferences.putUInt("highscore", highscore);
    preferences.end();
    wb32_fillRect(175, 120, 50, 15, wbBLACK);
    wb32_drawString(itoa(highscore, tmpbuf, 10), 175, 120, 1, 2);
  }
}

int check_line()
{
  int pt;
  for(i=19; i>0; i--) {
    for(j=0, pt=0; j<10; j++) if (board[i][j]!=0) pt++; else break;
    if (pt==10) {
      for(j=i; j>0; j--)
        for(k=0; k<10; k++) board[j][k]=board[j-1][k];
      return 1;
    }
  }
  return 0;
}

void go_next_blocks()
{
  pn=0;
  while(check_line()) { // check all stacked line
    cline++;
    pts+=20*(2^pn);
    pn++;
  }
  sfxn=pn+1; sfxc=0;
  ctype=nctype;
  nctype=random(7);
  pts+=10;
  cx = 4; cy = 0;
  rot=random(4);
  draw_blocks(cx, cy, ctype+1);
  draw_preview();
  update_score();
}

void init_blocks()
{
  clear_board();
  cline=0; level=0; pts=0;
  fall_limit = stage_limit = (level<=12?13-level:1);
  fall_time=0;
  cx=4; cy=0; 
  rot=random(4); ctype=random(7); nctype=random(7);
  draw_blocks(cx, cy, ctype+1);
  draw_preview();
  update_score();
}

void blocks_fall()
{
  if (++fall_time > fall_limit) {
    for(i=0; i<4; i++) {
      get_pos(pos=(cy+1)*10 + cx + blk_t[ctype*4+rot][i]);
      if (pos < 200) board[py-1][px]=0;
      else { // at bottom
        draw_blocks(cx, cy, ctype+1);
        go_next_blocks();
        return;
      }
    }
    cy++;
    for(i=0; i<4; i++) {
      get_pos(cy*10 + cx + blk_t[ctype*4+rot][i]);
      if (board[py][px]!=0) {
        draw_blocks(cx, cy-1, ctype+1);
        if (cy==1) {
          // game over, and restart
          sfxn=6; sfxc=0;
          wb32_fillRect(17, 100, 127, 50, wbRED);
          wb32_fillRect(22, 105, 117, 40, wbBLACK);
          wb32_drawString("Game Over", 32, 117, 2, 2);
          delay(2000); // wait 2 secs to prevent user's extra keys
          wb32_fillRect(16, 240, 128, 16, wbBLACK);
          wb32_drawString("Press a key to restart!", 21, 244, 2, 1);
          while(getkey()!=0xff); while(getkey()==0xff);
          while(getkey()!=0xff); 
          sfxn=7; sfxc=0;
          wb32_fillRect(20, 100, 130, 50, wbBLACK);
          init_blocks();
          return;
        }
        go_next_blocks();
        return;
      }
    }
    draw_blocks(cx, cy, ctype+1);
    fall_time=0;
  }
}

void check_left()
{
  for(i=0; i<4; i++) {
    get_pos(cy*10 + cx-1 + blk_t[ctype*4+rot][i]);
    if (px>cx+4) return; //left bound
  }
  draw_blocks(cx, cy, 0);
  for(i=0; i<4; i++) {
    get_pos(cy*10 + cx-1 + blk_t[ctype*4+rot][i]);
    if (board[py][px]!=0) {
      draw_blocks(cx, cy, ctype+1);
      return;
    }
  }
  cx--;
  draw_blocks(cx, cy, ctype+1);
}

void check_right()
{
  for(i=0; i<4; i++) {
    get_pos(cy*10 + cx+1 + blk_t[ctype*4+rot][i]);
    if (px>cx-1) {
      board[py][px-1]=0;
    } else {
      draw_blocks(cx, cy, ctype+1);
      return;
    }
  }
  for(i=0; i<4; i++) {
    get_pos(cy*10 + cx+1 + blk_t[ctype*4+rot][i]);
    if (board[py][px]!=0) {
      draw_blocks(cx, cy, ctype+1);
      return;
    }
  }
  cx++;
  draw_blocks(cx, cy, ctype+1);
}

int check_rotate()
{
  for(i=0; i<4; i++) {
    get_pos(cy*10 + cx + blk_t[ctype*4+rot][i]);
    if (pos<200) {
      if (board[py][px]!=0) return -2;
      if (px < cx) {
        cx--; return -1;
      }
      if (px > cx+4) {
        cx++; return -1;
      }
    }
  }
  return 0;
}

uint16_t sfx[8][8]={
  {0},
  {262, 0, 0, 0, 0, 0, 0, 0},
  {262, 330, 0, 0, 0, 0, 0, 0},
  {262, 330, 392, 0, 0, 0, 0, 0},
  {262, 330, 392, 523, 660, 0, 0, 0},
  {262, 330, 392, 523, 660, 784, 1047, 0},
  {262, 392, 330, 262, 16, 165, 131, 0},
  {1047,0,0,0,0,0,0,0}
};

void sfx_engine() // multi-tasking sound effect engine
{
  if (sfx_on) {
    if ((++task_c%3)==0) {
      if (sfxn!=0) {
        freq=sfx[sfxn][sfxc];
        if (freq) {
          ledcSetup(1, freq, 8);
          ledcWrite(1, 30);
        } else ledcWrite(1, 0);
        sfxc++;
        if (sfxc>7) sfxn=0;
      } 
    }
  }
}

hw_timer_t * _timer = NULL;  // the esp32 timer
void ticker_setup() // multitasking ticker engine
{
  wb32_tickerInit(30000, sfx_engine);
  ledcAttachPin(25, 1);
  ledcSetup(1, 200, 8);
  sfx_on=1;
  sfxn=0;
  task_c=0;
}

void init_preferences()
{
  preferences.begin("tetris",false);
  highscore=preferences.getUInt("highscore",1);
  if (highscore==1) { // first run
    preferences.putUInt("highscore", 0);
    highscore=0;
  }
  if ((getkey()&0x90)==0) preferences.putUInt("highscore", 0);
  preferences.end();
}

void init_game()
{
  init_keys();
  wb32_drawFastHLine(0,319,160, wbRED);
  wb32_drawFastVLine(0,0,320, wbRED);
  wb32_drawFastVLine(160,0,320, wbRED);
  init_blocks();
  wb32_setTextColor(wbYELLOW, wbYELLOW);
  wb32_drawString("Best", 170, 95, 2, 2);
  wb32_drawString("Score", 170, 150, 2, 2);
  wb32_drawString("Lines", 170, 205, 2, 2);
  wb32_drawString("Level", 170, 260, 2, 2);
  wb32_drawString("Boytris", 165, 2, 2, 2);
  wb32_setTextColor(wbGREEN, wbGREEN);
  wb32_drawString("(C) 2017 WiFiBoy.Org", 165, 310, 1, 1);
  wb32_drawString(itoa(highscore, tmpbuf, 10), 175, 120, 1, 2);
}

void check_downkey()
{
  if ((((key&0x10)==0)&&last_key==0)) { // down
    fall_limit=1; 
    last_key=5;
  } else if (last_key==5) { // down release
    fall_limit=stage_limit; 
    last_key=0;
  }
}

void check_leftkey()
{
  if (((key&0x80)==0)) { // repeatable key
    last_key=1;
    if (kc++%6==0) check_left(); // move delay = 6
    fall_limit=stage_limit;
  } else if (((key&0x80)==0x80)&&(last_key==1)) { // left release
    last_key=0; 
    kc=0;
  }
}

void check_rightkey()
{
  if (((key&0x40)==0)) { // repeatable key
    last_key=2;
    if (kc++%6==0) check_right(); // move delay = 6
    fall_limit=stage_limit; 
  } else if (((key&0x40)==0x40)&&(last_key==2)) { // left release
    last_key=0; 
    kc=0;
  }
}

void check_selectkey()
{
  if (((key&0x2)==0)&&(last_key==0)) { // repeatable key
    last_key=6;
    if (sfx_on==0) { 
      wb32_fillRect(230, 290, 8, 8, wbBLACK); 
      sfxn=0; 
    } else {
      wb32_setTextColor(wbRED, wbRED);
      wb32_drawString("X", 230, 290, 1, 1);
    }
    sfx_on = 1 - sfx_on;
  } else if (((key&0x2)==0x2)&&(last_key==6)) { // left release
    last_key=0; 
  }
}

void check_rotrkey()
{
  if (((key&0x04)==0)){ 
    last_key=3;
    if (kc++ % 15 == 0) { // rotate delay = 15
      draw_blocks(cx, cy, 0); // clean current blocks
      if (++rot>3) rot=0;
      while(1) { // dangerous!
        ret = check_rotate();
        if (ret==-2) {
          if (++rot>3) rot=0;
        } else if (ret==0) break; // pick next legal rotation
      }
      draw_blocks(cx, cy, ctype+1); // draw rotated blocks
    }
  } else if (((key&0x04)==0x04)&&(last_key==3)) {
    last_key=0;
    kc=0;
  }
}

void check_rotlkey()
{ 
  if (((key&0x08)==0)) { 
    last_key=4;
    if (kc++ % 15 == 0) { // rotate delay = 15
      draw_blocks(cx, cy, 0); 
      if (--rot<0) rot=3;
      while(1) { // dangerous!
        ret = check_rotate();
        if (ret==-2) {
          if (--rot<0) rot=3;
        } else if (ret==0) break; 
      }
      draw_blocks(cx, cy, ctype+1);
    }
  } else if (((key&0x08)==0x08)&&(last_key==4)) {
    last_key=0;
    kc=0;
  }
}

void setup() 
{
  wb32_init();
  init_preferences();
  ticker_setup();
  init_game();
}

void loop() // game loop (40~50fps)
{ 
  getkey();
  check_downkey();
  check_leftkey();
  check_rightkey();
  check_rotrkey();
  check_rotlkey();
  check_selectkey();
  draw_board();
  blocks_fall();
  delay(20);
}
